//#define IS_SSE2
// Copyright (c) 2002 Tom Barry.  All rights reserved.
//      trbarry@trbarry.com
// Requires Avisynth source code to compile for Avisynth
// Avisynth Copyright 2000 Ben Rudiak-Gould.
//      http://www.math.berkeley.edu/~benrg/avisynth.html
/////////////////////////////////////////////////////////////////////////////
//
//	This file is subject to the terms of the GNU General Public License as
//	published by the Free Software Foundation.  A copy of this license is
//	included with this software distribution in the file COPYING.  If you
//	do not have a copy, you may obtain a copy by writing to the Free
//	Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
//	This software is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details
//  
//  Also, this program is "Philanthropy-Ware".  That is, if you like it and 
//  feel the need to reward or inspire the author then please feel free (but
//  not obligated) to consider joining or donating to the Electronic Frontier
//  Foundation. This will help keep cyber space free of barbed wire and bullsh*t.  
//
/////////////////////////////////////////////////////////////////////////////
// Change Log
//
// Date          Developer             Changes
//
// 29 Jun 2003   0.0.1.7    Tom Barry	       Fix YV12 crash bug when TF=0
        
// 17 Jun 2003   0.0.1.6    Tom Barry	       Fix SearchEffortW for YUY2
//											   Change various "2" to BPP for YV12 (bits/pixel

// 21 Jan 2003   0.0.1.5    Tom Barry	       Same source compiled without Optimize 

// 18 Jan 2003   0.0.1.4    Tom Barry	       Use AvisynthPluginInit2 

// 12 Nov 2002   0.0.1.3    Tom Barry	       Try again to make YV12 support 

// 15 Apr 2002   0.0.1.0    Tom Barry	       Create TomsMoComp Deinterlacer 
//
/////////////////////////////////////////////////////////////////////////////
// CVS Log
//
//
/////////////////////////////////////////////////////////////////////////////
/*

From the Readme_TomsMoComp.txt file:

        TomsMoComp - Motion Compensating Deinterlace


        TomsMoComp.dll is an Avisynth filter that uses motion compensation and adaptive
		processing to deinterlace video source.

        Just unzip the contents into your Avisynth directory, or somewhere. As the script shows,
        I made a subdirectory under Avisynth just to keep it separate.
		avs, one of the scripts I was testing:

        LoadPlugin("d:\AVISynth\TomsMoComp\Debug\TomsMoComp.dll")
        clip = AVISource("c:\vcr\bikes.avi")
        return clip.TomsMoComp(...)

        It specifies the file spec (change yours) and asks for TopFirst and SearchEffort=10 
        to be turned on. I've so far tested it only 
        with Avisynth/VirtualDub. 


        TomsMoComp Parm list:

          return clip.TomsMoComp(TopFirst, SearchEffort, Vertical_Filter)

        All the values are integer, 0=no, 1=yes:
        
          TopFirst - assume the top field, lines 0,2,4,... should be displayed first. 
	        The default is the supposedly more common BottomFirst (not for me).

		SearchEffort - determines how much effort (cpu time) will be used to find 
		    moved pixels. Currently numbers from 0-45 with 0 being practially
			just a smarter bob.

        Known issues and limitation:
        1) 	Assumes YUV (YUY2) Frame Based input. Use an AVIsynth function to convert first if
	        needed.        


        Temporary file locations:

        For now, see www.trbarry.com/TomsMoComp.zip

		THIS PROGRAM IS STILL NEW AND EXPERIMENTAL!

  Please send comments to Tom Barry (trbarry@trbarry.com)


*/
//#include "internal.h"
#include "windows.h"
#include "avisynth.h"

#include <malloc.h>
#include <string.h>

#include "TomsMoComp.h"

extern  "C" BOOL AvisynthTomsMoComp_SSE(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);

extern  "C" BOOL AvisynthTomsMoComp_SSE_VF(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);

extern  "C" BOOL AvisynthTomsMoComp_3DNOW(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);

extern  "C" BOOL AvisynthTomsMoComp_3DNOW_VF(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);

extern  "C" BOOL AvisynthTomsMoComp_MMX(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);

extern  "C" BOOL AvisynthTomsMoComp_MMX_VF(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);

BOOL	Avisynth_DblResizeH(int src_pitch, int dst_pitch,
			int rowsize, const BYTE* srcp, BYTE* dstp,
			int FldHeight);					// go H expand lines into even output lines

// again for YV12 support
extern  "C" BOOL AvisynthTomsMoComp_SSE_YV12(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);


extern  "C" BOOL AvisynthTomsMoComp_SSE_VF_YV12(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);


extern  "C" BOOL AvisynthTomsMoComp_3DNOW_YV12(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);

extern  "C" BOOL AvisynthTomsMoComp_3DNOW_VF_YV12(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);

extern  "C" BOOL AvisynthTomsMoComp_MMX_YV12(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);

extern  "C" BOOL AvisynthTomsMoComp_MMX_VF_YV12(int SearchEffort, int _src_pitch, int _dst_pitch,
		int _rowsize, const BYTE* _pWeaveSrc, const BYTE* _pWeaveSrcP,
		BYTE* _pWeaveDest, BOOL _TopFirst, const BYTE* _pCopySrc,
		const BYTE* _pCopySrcP, int _FldHeight);




class TomsMoComp : public GenericVideoFilter {
    BOOL TopFirst;
    BOOL Use_Vertical_Filter;
    int  Max_Frames;
    int  SearchEffort;
    int  SearchEffortW;
    int  PrevFrame;
    int  PrevInFrame;
    int  dropFrame;
	int  useFrame;
    const unsigned char* srcp;
    const unsigned char* prev_srcp;

    PVideoFrame src;
    PVideoFrame dst;
    PVideoFrame prevSrc;
    unsigned char* dstp;

	const BYTE* pWeaveSrc;
	const BYTE* pWeaveSrcP;
	BYTE* pWeaveDest;
	const BYTE* pCopySrc;
	const BYTE* pCopySrcP;
	BYTE* pCopyDest;
	BYTE* pWorkArea;

	BYTE* oldBuff;
	BYTE* newBuff;
	bool SSE2enabled;
	bool SSEMMXenabled;
	bool _3DNOWenabled;

	// YUY2 only
    int src_pitch;
    int dst_pitch;
    int rowsize;
    int RowPixels;
    int height;
    int FldHeight;

	int __stdcall Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch);

public:
    TomsMoComp(PClip _child, BOOL _TopFirst, int _Search_Effort, int _Use_Vertical_Filter,
		IScriptEnvironment* env) : GenericVideoFilter(_child)
    {
        TopFirst = _TopFirst;
        Use_Vertical_Filter = _Use_Vertical_Filter;
        SearchEffort = _Search_Effort;
        Max_Frames = vi.num_frames;
        PrevFrame = -2;                 // last frame processed, if any
        PrevInFrame = -2;               // last frame requested, if any
        dropFrame = 1;                  // next frame to drop
		SSE2enabled = (env->GetCPUFlags() & CPUF_SSE2) !=0;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;
		_3DNOWenabled = (env->GetCPUFlags() & CPUF_3DNOW) !=0;


	    if (!vi.IsYUY2() && ! vi.IsYV12())
		{
			env->ThrowError("TomsMoComp: Supports YUY2 and YV12 color formats only");
		}
	    if (vi.width % 4)
		{
			env->ThrowError("TomsMoComp: Horizontal pixels must divide by 4");
		}
		
		if (SearchEffort == -1)
		{
			vi.height = 2 * vi.height;
			TopFirst = 0;
		}		
		
		if (SearchEffort == -2)
		{
			vi.width = 2 * vi.width;
			vi.height = 2 * vi.height;
			TopFirst = 0;
			if (Use_Vertical_Filter)
			{
				pWorkArea = (BYTE*) malloc(vi.width*vi.height*2+128); 
			}
		}	
//		vi.pixel_type &= ~VideoInfo::CS_FIELDBASED;  // say is now frame based after deinterlace
    }

    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);
/*    int __stdcall ProcessFrame(PVideoFrame dst, unsigned char* dstp, 
		BOOL WriteFrame, IScriptEnvironment* env); 
	int __stdcall ProcessFrame_YV12(PVideoFrame dst, unsigned char* dstp, 
		BOOL WriteFrame, IScriptEnvironment* env);
 */
	int __stdcall ProcessFrame(unsigned char* dstp, 
		BOOL WriteFrame, IScriptEnvironment* env); 
	int __stdcall ProcessFrame_YV12(unsigned char* dstp, 
		BOOL WriteFrame, IScriptEnvironment* env);

};


PVideoFrame __stdcall TomsMoComp::GetFrame(int inFrame, IScriptEnvironment* env) 
{
int i;

	useFrame = inFrame;
	src = child->GetFrame(useFrame, env);
	dst = env->NewVideoFrame(vi);

	if (TopFirst == -1)
		{
			TopFirst = (child->GetParity(1))
				? 1
				: 0;
		}

    if (useFrame > Max_Frames-1 || useFrame < 0)
    {
        return dst;
    }

	SearchEffortW = 0;
	
	if (useFrame)
	{
		SearchEffortW = SearchEffort;
		prevSrc = child->GetFrame(useFrame-1, env);
		pWeaveSrcP = TopFirst
			? prevSrc->GetReadPtr() + src_pitch
			: prevSrc->GetReadPtr();
		pCopySrcP = TopFirst
			? prevSrc->GetReadPtr() + src_pitch
			: prevSrc->GetReadPtr();
	}

	if (vi.IsYUY2())
	{
	    srcp = src->GetReadPtr();
		dstp = dst->GetWritePtr();
		src_pitch = src->GetPitch();
		dst_pitch = dst->GetPitch();
		rowsize = dst->GetRowSize();
		RowPixels = rowsize >> 1;
		height = dst->GetHeight();
		FldHeight = height / 2;
		if (useFrame)
		{
			pWeaveSrcP = TopFirst
				? prevSrc->GetReadPtr() + src_pitch
				: prevSrc->GetReadPtr();
			pCopySrcP = TopFirst
				? prevSrc->GetReadPtr() + src_pitch
				: prevSrc->GetReadPtr();
		}
       
		i = ProcessFrame(dstp, true, env);       // table and then write the frame
    }
	else
	{
		// For YV12 we are going to cheat and deinterlace each plane separately as if it was
		// the whole frame.  This is a quick hack, but probably will be okay.
		// First, the Y plane, pretending it is the whole frame
		srcp = src->GetReadPtr(PLANAR_Y);
		dstp = dst->GetWritePtr(PLANAR_Y);
		src_pitch = src->GetPitch(PLANAR_Y);
		dst_pitch = dst->GetPitch(PLANAR_Y);
		rowsize = dst->GetRowSize(PLANAR_Y);   // Could also be PLANAR_Y_ALIGNED which would return a mod16 row_size   
		height = dst->GetHeight(PLANAR_Y);
		FldHeight = height / 2;
		if (useFrame)
		{
			pWeaveSrcP = TopFirst
				? prevSrc->GetReadPtr(PLANAR_Y) + src_pitch
				: prevSrc->GetReadPtr(PLANAR_Y);
			pCopySrcP = TopFirst
				? prevSrc->GetReadPtr(PLANAR_Y) + src_pitch
				: prevSrc->GetReadPtr(PLANAR_Y);
		}

		i = ProcessFrame_YV12(dstp, true, env);       // table and then write the frame

		// Next, the U plane
		srcp = src->GetReadPtr(PLANAR_U);
		dstp = dst->GetWritePtr(PLANAR_U);
		src_pitch = src->GetPitch(PLANAR_U);
		dst_pitch = dst->GetPitch(PLANAR_U);
		rowsize = dst->GetRowSize(PLANAR_U);  // Could also be PLANAR_U_ALIGNED which would return a mod8 row_size   
		height = dst->GetHeight(PLANAR_U);
		FldHeight = height / 2;
		if (useFrame)
		{
			pWeaveSrcP = TopFirst
				? prevSrc->GetReadPtr(PLANAR_U) + src_pitch
				: prevSrc->GetReadPtr(PLANAR_U);
			pCopySrcP = TopFirst
				? prevSrc->GetReadPtr(PLANAR_U) + src_pitch
				: prevSrc->GetReadPtr(PLANAR_U);
		}
		i = ProcessFrame_YV12(dstp, true, env);       // table and then write the frame

		// And the V plane, same sizes as U, different locations
		srcp = src->GetReadPtr(PLANAR_V);   
		dstp = dst->GetWritePtr(PLANAR_V);   
		if (useFrame)
		{
			pWeaveSrcP = TopFirst
				? prevSrc->GetReadPtr(PLANAR_V) + src_pitch
				: prevSrc->GetReadPtr(PLANAR_V);
			pCopySrcP = TopFirst
				? prevSrc->GetReadPtr(PLANAR_V) + src_pitch
				: prevSrc->GetReadPtr(PLANAR_V);
		}
		i = ProcessFrame_YV12(dstp, true, env);       // table and then write the frame

	}
	PrevFrame = useFrame;
    PrevInFrame = inFrame;
	
	__asm emms;

    return dst;
}

#define IS_SSE
#define SSE_TYPE SSE


BOOL DblResizeH_1(int rowsize, const BYTE* srcp, BYTE* dstp)	
{
int ct = (rowsize-8) >> 3;
		__asm
		{
		// Loop general reg usage
		//
		// ecx loop ctr
		// edi - dest
		// esi - src pixels

		// now loop and get the middle qwords
		mov		edi, dstp
		mov		esi, srcp  
		mov		ecx, ct
		
LoopQ:	
// just use the current line, first luma
		movq	mm0, qword ptr[esi]		// orig luma, in register as VYUYVYUY
		movq	mm2, mm0				// keep a copy while we got it
		movq	mm6, qword ptr[esi+2]	// luma 2 bytes to right
		pavgb   mm6, mm0				// 3 pixels of valid new luma averag, plus garbage
		pand	mm0, YMask				// keep only old luma, 0Y0Y0Y0Y
		psllw	mm6, 8					// now only new luma as y0y0y0y0
		por		mm6, mm0				// all luma yYyYyYyY

// now  chroma bytes
		movq	mm1, mm2
		pavgb	mm1, qword ptr[esi+4]	// chroma 4 bytes to right
		psrlw	mm2, 8					// only orig chroma as 0V0U0V0U
		packuswb mm2, mm2				// now xxxxVUVU
		psrlw	mm1, 8					// only new chroma as 0?0?0v0u
		packuswb mm1, mm1				// now new chroma bytes xxxx??vu
		punpcklwd mm2,mm1				// ??VUvuVU, merged old and new chroma

// merge final luma & chroma bytes

		movq	mm7, mm6				// copy of luma, ?YyYyYyY
		punpcklbw mm6, mm2				// low bytes vyuYVyUY
		punpckhbw mm7, mm2				// hi bytes ???YVyUY


		V_MOVNTQ (qword ptr[edi], mm6)	// store low qword
		V_MOVNTQ (qword ptr[edi+8], mm7) // store high qword

		lea		esi, [esi+8]		     // bump src pointer
		lea		edi, [edi+16]		// bump dest pointer
		dec		ecx
		jg		LoopQ				// y	

// now expand the last 8 into 16 bytes
		movq	mm0, qword ptr[esi]		// orig luma, in register as VYUYVYUY
		movq	mm2, mm0				// keep a copy while we got it
		movq	mm6, mm0				// and another to work on now
		
		psrlq	mm6, 16					// pretend most of it came from +2 offs (remem revers order)
		pavgb   mm6, mm0				// 3 pixels of valid new luma averag, plus garbage
		pand	mm0, YMask				// keep only old luma, 0Y0Y0Y0Y
		psllw	mm6, 8					// now only new luma as ?0y0y0y0
		por		mm6, mm0				// all luma ?YyYyYyY

// now final chroma bytes
		movq	mm1, mm2
		psrlq	mm1, 32					// pretend most of it came from +4 offs (remem revers order)
		pavgb   mm1, mm2				// 2 pixels of valid new chroma averag, plus garbage
		psrlw	mm2, 8					// only orig chroma as 0V0U0V0U
		packuswb mm2, mm2				// now xxxxVUVU
		psrlw	mm1, 8					// only new chroma as 0?0?0v0u
		packuswb mm1, mm1				// now new chroma bytes xxxx??vu
		punpcklwd mm2,mm1				// ??VUvuVU, merged old and new chroma

// merge final luma & chroma bytes

		movq	mm7, mm6				// copy of luma, ?YyYyYyY
		punpcklbw mm6, mm2				// low bytes vyuYVyUY
		punpckhbw mm7, mm2				// hi bytes ???YVyUY

		movq qword ptr[edi], mm6	   // store low qword
		movq qword ptr[edi+8], mm7     // store high qword
		
		mov al, byte ptr[esi+5]    // last U source byte
		mov byte ptr[edi+13], al
		mov	ax, word ptr[esi+6]	   // last 2 YV source bytes
		mov word ptr[edi+14], ax
		}	
		return 0;
}




BOOL	Avisynth_DblResizeH(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp,
			int FldHeight)					// go H expand lines into even output lines
{
__int64 SaveLuma;

const BYTE* pSrc = srcp;
BYTE* pDest = dstp + 2*dst_pit;
int y;
int ct = (rowsize-8) >> 3;

	DblResizeH_1(rowsize, srcp, dstp);		// expand top line
	DblResizeH_1(rowsize, srcp + src_pit, dstp + dst_pit);		// expand 2nd line
	DblResizeH_1(rowsize, srcp+(FldHeight-2) * src_pit,
		dstp + (FldHeight-2) * dst_pit);		// expand next to last line
	DblResizeH_1(rowsize, srcp+(FldHeight-1) * src_pit,
		dstp + (FldHeight-1) * dst_pit);		// expand last line
	
	for (y=1; y <= FldHeight-4; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax -  
		// ebx - src 2 lines above curr 
		// ecx loop ctr
		// edx - dst_pit
		// edi - dest
		// esi - src pixels, 1 line up

		// now loop and get the middle qwords
		mov		edi, pDest
		mov		ebx, pSrc   
		mov		ecx, ct 
		mov		edx, src_pit 

LoopQ:	
		// Assume our pixels are layed out as follows with x the calc'd value
		// and the other pixels are from the current field, top, curr, and lower lines.
		//  
		//		  i   j 		2 lines above 
		//		  a   b 		above line
		//        c x d			curr line
		//        e   f 		next line
		//        m   n 		2 lines below
		//
		// we calc the middle x value as:
		//		x = either avg(a,f), avg(c,d), avg(b,e)
		 
		// selected for the	smallest of SAD(a,f), SAD(c,d), or SAD(b,e), etc.
		// where SAD() = sum of absolute diff
		//
		// we do this first for luma. For now we will just horizontally average chroma
		
		lea		esi, [ebx+edx]

		// i,n
		movq    mm0, qword ptr[ebx]			// value i from top left		
		movq    mm1, qword ptr[ebx+4*edx+2]	// value n from bottom right			
		movq	mm6, mm0					
//		pavgb	mm6, mm1					// avg(i,n), also best so far
		V_PAVGB (mm6, mm1, mm7, ShiftMask)	// avg(i,n), also best so far
		movq	mm7, mm0
		psubusb	mm7, mm1
		psubusb mm1, mm0
		por		mm7, mm1					// abs diff (SAD), also best so far

		// m,j
		movq    mm0, qword ptr[ebx+2]		// value j from top right		
		movq    mm1, qword ptr[ebx+4*edx]	// value m from bottom left			
		ISBOBBETTER

		// a,f
		movq    mm0, qword ptr[esi]			// value a from top left		
		movq    mm1, qword ptr[esi+2*edx+2]	// value f from bottom right			
		ISBOBBETTER


		// e,b
		movq    mm0, qword ptr[esi+2]		// value b from top right		
		movq    mm1, qword ptr[esi+2*edx]	// value e from bottom left			
		ISBOBBETTER

		// c,d
		movq    mm0, qword ptr[esi+edx]	// value c from left 		
		movq    mm1, qword ptr[esi+edx+2]	// value d from right

		// We will also clip what we've got so far to avoid artifacts

		movq	mm2, mm0
//		pminub	mm2, mm1
		V_PMINUB (mm2, mm1, mm4)
//		pmaxub	mm6, mm2			// clip our current results so far to be above this
		V_PMAXUB (mm6, mm2)
		movq	mm2, mm0
		V_PMAXUB (mm2, mm1)
//		pminub	mm6, mm2			// clip our current results so far to be below this
		V_PMINUB (mm6, mm2, mm4)

		ISBOBBETTER
		
// we have now create 4 new luma values for pixel offsets 1,3,5,7. But we need to merge these
// with luma values for offsets 0,2,4,6 and some chroma info. First the other luma.
		movq	mm0, qword ptr[esi+edx] // orig luma, in register as VYUYVYUY
		movq	mm2, mm0				// keep a copy while we got it
		pand	mm0, YMask				// keep only old luma, 0Y0Y0Y0Y
		psllw	mm6, 8					// now only new luma as y0y0y0y0
		por		mm6, mm0				// all luma yYyYyYyY
		movq	SaveLuma, mm6			// and save it for later

// get chroma, but we only use 3 lines for chroma instead of 5


		// a,f - Chroma
		movq    mm0, qword ptr[esi]			// value a from top left		
		movq    mm1, qword ptr[esi+2*edx+4]	// value f from bottom right			
		movq	mm6, mm0					
//		pavgb	mm6, mm1					// avg(a,f), also best so far
		V_PAVGB (mm6, mm1, mm7, ShiftMask)	
		movq	mm7, mm0
		psubusb	mm7, mm1
		psubusb mm1, mm0
		por		mm7, mm1					// abs diff (SAD), also best so far

		// e,b - Chroma
		movq    mm0, qword ptr[esi+4]		// value b from top right		
		movq    mm1, qword ptr[esi+2*edx]	// value e from bottom left			
		ISBOBBETTER

		// c,d - Chroma
		movq    mm0, qword ptr[esi+edx]	// value c from left 		
		movq    mm1, qword ptr[esi+edx+4]	// value d from right

		// We will also clip what we've got so far to avoid artifacts
		movq	mm2, mm0
//		pminub	mm2, mm1
		V_PMINUB (mm2, mm1, mm4)
//		pmaxub	mm6, mm2			// clip our current results so far to be above this
		V_PMAXUB (mm6, mm2)
		movq	mm2, mm0
		V_PMAXUB (mm2, mm1)
//		pminub	mm6, mm2			// clip our current results so far to be below this
		V_PMINUB (mm6, mm2, mm4)

		ISBOBBETTER

		movq    mm2, qword ptr[esi+edx]	// another copy of orig Chroma from c
		psrlw	mm2, 8					// only orig chroma as 0V0U0V0U
		packuswb mm2, mm2				// now xxxxVUVU
		psrlw	mm6, 8					// only new chroma as 0v0u0v0u
		packuswb mm6, mm6				// now new chroma bytes xxxxvuvu
		punpcklwd mm2,mm6				// vuVUvuVU, merged old and new chroma

// merge luma & chroma
		movq	mm6, SaveLuma			// get luma again		
		movq	mm7, mm6				// copy of luma
		punpcklbw mm6, mm2				// low bytes vyuYVyUY
		punpckhbw mm7, mm2				// hi bytes vyuYVyUY

//		movntq	qword ptr[eax+edx], mm0
		V_MOVNTQ (qword ptr[edi], mm6)	// store low qword
		V_MOVNTQ (qword ptr[edi+8], mm7) // store high qword
		lea		ebx, [ebx+8]		     // bump src pointer
		lea		edi, [edi+16]		// bump dest pointer
		dec		ecx
		jg		LoopQ				// y	

// done with our horizontal loop but we have 8 more bytes of input to process
// we will be fairly lazy with this and for now just use the current line, first luma
		lea		esi, [ebx+edx]
		movq	mm0, qword ptr[esi+edx] // orig luma, in register as VYUYVYUY
		movq	mm2, mm0				// keep a copy while we got it
		movq	mm6, mm0				// and another to work on now
		
		psrlq	mm6, 16					// pretend most of it came from +2 offs (remem revers order)
		pavgb   mm6, mm0				// 3 pixels of valid new luma averag, plus garbage
		pand	mm0, YMask				// keep only old luma, 0Y0Y0Y0Y
		psllw	mm6, 8					// now only new luma as ?0y0y0y0
		por		mm6, mm0				// all luma ?YyYyYyY

// now final chroma bytes
		movq	mm1, mm2
		psrlq	mm1, 32					// pretend most of it came from +4 offs (remem revers order)
		pavgb   mm1, mm2				// 2 pixels of valid new chroma averag, plus garbage
		psrlw	mm2, 8					// only orig chroma as 0V0U0V0U
		packuswb mm2, mm2				// now xxxxVUVU
		psrlw	mm1, 8					// only new chroma as 0?0?0v0u
		packuswb mm1, mm1				// now new chroma bytes xxxx??vu
		punpcklwd mm2,mm1				// ??VUvuVU, merged old and new chroma

// merge final luma & chroma bytes

		movq	mm7, mm6				// copy of luma, ?YyYyYyY
		punpcklbw mm6, mm2				// low bytes vyuYVyUY
		punpckhbw mm7, mm2				// hi bytes ???YVyUY

		movq qword ptr[edi], mm6	   // store low qword
		movq qword ptr[edi+8], mm7     // store high qword
		
		mov al, byte ptr[esi+edx+5]    // last U source byte
		mov byte ptr[edi+13], al
		mov	ax, word ptr[esi+edx+6]	   // last 2 YV source bytes
		mov word ptr[edi+14], ax
		}	


		// adjust for next line
		pSrc  += src_pit;
		pDest += dst_pit;
	}
	
	return 0;


}
int __stdcall TomsMoComp::ProcessFrame(unsigned char* dstp, 
                                      BOOL WriteFrame, IScriptEnvironment* env) 
{

    BOOL retc = 0;
    if (SearchEffort == -2)				// note - DBL_Resize carried ast TopFirst == 0 hrtrtr
    {
		SearchEffortW = -1;				// pretend it is -1 after this
		pWeaveDest = dstp+dst_pitch;	// odd dest lines
		pCopyDest = dstp;				// even dest lines

		if (Use_Vertical_Filter)
		{
			Avisynth_DblResizeH(src_pitch, rowsize,
				rowsize/2, srcp, pWorkArea,
				FldHeight);					// go H expand lines into even output lines
			
			src_pitch = rowsize;			// pretend it is wider now, and in out buffer
			pWeaveSrc = pWorkArea;			// pretend source is here now
			pCopySrc = pWorkArea;			// "
			Fieldcopy(pCopyDest, pCopySrc, rowsize,				// copy top EVEN line
						1, dst_pitch, src_pitch);
		}
		else
		{
			Avisynth_DblResizeH(src_pitch, 2*dst_pitch,
				rowsize/2, srcp, dstp,
				FldHeight);					// go H expand lines into even output lines

			src_pitch = 2*dst_pitch;		// pretend it is wider now, and in out buffer
			pWeaveSrc = dstp;				// pretend source is here now
			pCopySrc = dstp;				// "
//			Fieldcopy(pWeaveDest, pCopySrc, rowsize,				// copy top ODD line
//					1, dst_pitch, src_pitch);
		}

		Fieldcopy(pWeaveDest + (height-2)*dst_pitch,			// copy bottom Odd line	
			pCopySrc+(FldHeight-1)*src_pitch, rowsize, 
					1, dst_pitch, src_pitch);

	}
    else if (SearchEffort == -1)				// note - DBL_Resize carried ast TopFirst == 0 hrtrtr
    {
		SearchEffortW = SearchEffort;
		pCopySrc = srcp;						// even src lines, 0,2,4..
		pWeaveSrc = srcp;						// odd src lines, 1,3,5..
		pCopyDest = dstp;						// even dest
		pWeaveDest = dstp+dst_pitch;			// odd dest
		if (Use_Vertical_Filter)
		{
			Fieldcopy(pCopyDest, pCopySrc, rowsize,				// copy top EVEN line
						1, dst_pitch, src_pitch);
		}
		else
		{
			Fieldcopy(pCopyDest, pCopySrc, rowsize,			// copy all EVEN lines (base 0)
					FldHeight, dst_pitch*2, src_pitch);		// I don't remember why???
		}
		Fieldcopy(pWeaveDest + (height-2)*dst_pitch,			// copy bottom Odd line	
			pCopySrc+(FldHeight-1)*src_pitch, rowsize, 
					1, dst_pitch, src_pitch);
	}
    else if (TopFirst)
    {
		pWeaveSrc = srcp+src_pitch;
		pCopySrc = srcp;
		pWeaveDest = dstp+dst_pitch;
		pCopyDest = dstp;
		Fieldcopy(pWeaveDest, pCopySrc, rowsize,				// bob this later?
					1, dst_pitch*2, src_pitch*2);
		Fieldcopy(pWeaveDest+(FldHeight-1)*dst_pitch*2,
			pCopySrc+(FldHeight-1)*src_pitch*2, rowsize, 
					1, dst_pitch*2, src_pitch*2);
		if (Use_Vertical_Filter)
		{
			Fieldcopy(pCopyDest, pCopySrc, rowsize, 
						1, dst_pitch*2, src_pitch*2);
			Fieldcopy(pCopyDest+(FldHeight-1)*dst_pitch*2,		// copy last bob line
						pCopySrc+(FldHeight-1)*src_pitch*2, rowsize, 
						1, dst_pitch*2, src_pitch*2);
		}
		else
		{
			Fieldcopy(pCopyDest, pCopySrc, rowsize, 
						FldHeight, dst_pitch*2, src_pitch*2);
		}
	}
    else             // not Top First
	{
		pWeaveSrc = srcp;
		pCopySrc = srcp + src_pitch;
		pWeaveDest = dstp;
		pCopyDest = dstp + dst_pitch;
		Fieldcopy(pWeaveDest, pCopySrc, rowsize, 
					1, dst_pitch*2, src_pitch*2);
		Fieldcopy(pWeaveDest+(FldHeight-1)*dst_pitch*2,			// bob this later
			pCopySrc+(FldHeight-1)*src_pitch*2, rowsize, 
					1, dst_pitch*2, src_pitch*2);
		if (Use_Vertical_Filter)
		{	
			Fieldcopy(pCopyDest, pCopySrc, rowsize, 
						1, dst_pitch*2, src_pitch*2);			// copy first bob line
			Fieldcopy(pCopyDest+(FldHeight-1)*dst_pitch*2,
				pCopySrc+(FldHeight-1)*src_pitch*2, rowsize,	// copy last bob line
						1, dst_pitch*2, src_pitch*2);
			pWeaveDest += dst_pitch;		// bug, adj for poor planning
		}
		else
		{
			Fieldcopy(pCopyDest, pCopySrc, rowsize, 
						FldHeight, dst_pitch*2, src_pitch*2);
		}
	}
	if (Use_Vertical_Filter)
	{
		if (SSEMMXenabled)
		{
			AvisynthTomsMoComp_SSE_VF(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
		else if (_3DNOWenabled)
		{
			AvisynthTomsMoComp_3DNOW_VF(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
		else
		{
			AvisynthTomsMoComp_MMX_VF(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
	}
	else
	{
		if (SSEMMXenabled)
		{
			AvisynthTomsMoComp_SSE(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
		else if (_3DNOWenabled)
		{
			AvisynthTomsMoComp_3DNOW(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
		else
		{
			AvisynthTomsMoComp_MMX(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
	}
	return 0;
}

int __stdcall TomsMoComp::ProcessFrame_YV12(unsigned char* dstp, 
                                      BOOL WriteFrame, IScriptEnvironment* env) 
{

    BOOL retc = 0;
/* not yet >>>>>>>>>
    if (SearchEffort == -2)				// note - DBL_Resize carried ast TopFirst == 0 hrtrtr
    {
		SearchEffortW = -1;				// pretend it is -1 after this
		pWeaveDest = dstp+dst_pitch;	// odd dest lines
		pCopyDest = dstp;				// even dest lines

		if (Use_Vertical_Filter)
		{
			Avisynth_DblResizeH(src_pitch, rowsize,
				rowsize/2, srcp, pWorkArea,
				FldHeight);					// go H expand lines into even output lines
			
			src_pitch = rowsize;			// pretend it is wider now, and in out buffer
			pWeaveSrc = pWorkArea;			// pretend source is here now
			pCopySrc = pWorkArea;			// "
			Fieldcopy(pCopyDest, pCopySrc, rowsize,				// copy top EVEN line
						1, dst_pitch, src_pitch);
		}
		else
		{
			Avisynth_DblResizeH(src_pitch, 2*dst_pitch,
				rowsize/2, srcp, dstp,
				FldHeight);					// go H expand lines into even output lines

			src_pitch = 2*dst_pitch;		// pretend it is wider now, and in out buffer
			pWeaveSrc = dstp;				// pretend source is here now
			pCopySrc = dstp;				// "
//			Fieldcopy(pWeaveDest, pCopySrc, rowsize,				// copy top ODD line
//					1, dst_pitch, src_pitch);
		}

		Fieldcopy(pWeaveDest + (height-2)*dst_pitch,			// copy bottom Odd line	
			pCopySrc+(FldHeight-1)*src_pitch, rowsize, 
					1, dst_pitch, src_pitch);

	}
    else */
	if (SearchEffort < -1)
		SearchEffortW = -1;		//>>>>> temp

	if (SearchEffort == -1)				// note - DBL_Resize carried ast TopFirst == 0 hrtrtr
    {
		SearchEffortW = SearchEffort;
		pCopySrc = srcp;						// even src lines, 0,2,4..
		pWeaveSrc = srcp;						// odd src lines, 1,3,5..
		pCopyDest = dstp;						// even dest
		pWeaveDest = dstp+dst_pitch;			// odd dest
		if (Use_Vertical_Filter)
		{
			Fieldcopy(pCopyDest, pCopySrc, rowsize,				// copy top EVEN line
						1, dst_pitch, src_pitch);
		}
		else
		{
			Fieldcopy(pCopyDest, pCopySrc, rowsize,			// copy all EVEN lines (base 0)
					FldHeight, dst_pitch*2, src_pitch);		// I don't remember why???
		}
		Fieldcopy(pWeaveDest + (height-2)*dst_pitch,			// copy bottom Odd line	
			pCopySrc+(FldHeight-1)*src_pitch, rowsize, 
					1, dst_pitch, src_pitch);
	}
	else 


	if (TopFirst)
	{
		pWeaveSrc = srcp+src_pitch;
		pCopySrc = srcp;
		pWeaveDest = dstp+dst_pitch;
		pCopyDest = dstp;
		Fieldcopy(pWeaveDest, pCopySrc, rowsize,				// bob this later?
					1, dst_pitch*2, src_pitch*2);
		Fieldcopy(pWeaveDest+(FldHeight-1)*dst_pitch*2,
			pCopySrc+(FldHeight-1)*src_pitch*2, rowsize, 
					1, dst_pitch*2, src_pitch*2);
		if (Use_Vertical_Filter)
		{
			Fieldcopy(pCopyDest, pCopySrc, rowsize, 
						1, dst_pitch*2, src_pitch*2);
			Fieldcopy(pCopyDest+(FldHeight-1)*dst_pitch*2,		// copy last bob line
						pCopySrc+(FldHeight-1)*src_pitch*2, rowsize, 
						1, dst_pitch*2, src_pitch*2);
		}
		else
		{
			Fieldcopy(pCopyDest, pCopySrc, rowsize, 
						FldHeight, dst_pitch*2, src_pitch*2);
		}
	}
    else             // not Top First
	{
		pWeaveSrc = srcp;
		pCopySrc = srcp + src_pitch;
		pWeaveDest = dstp;
		pCopyDest = dstp + dst_pitch;
		Fieldcopy(pWeaveDest, pCopySrc, rowsize, 
					1, dst_pitch*2, src_pitch*2);
		Fieldcopy(pWeaveDest+(FldHeight-1)*dst_pitch*2,			// bob this later
			pCopySrc+(FldHeight-1)*src_pitch*2, rowsize, 
					1, dst_pitch*2, src_pitch*2);
		if (Use_Vertical_Filter)
		{	
			Fieldcopy(pCopyDest, pCopySrc, rowsize, 
						1, dst_pitch*2, src_pitch*2);			// copy first bob line
			Fieldcopy(pCopyDest+(FldHeight-1)*dst_pitch*2,
				pCopySrc+(FldHeight-1)*src_pitch*2, rowsize,	// copy last bob line
						1, dst_pitch*2, src_pitch*2);
			pWeaveDest += dst_pitch;		// bug, adj for poor planning
		}
		else
		{
			Fieldcopy(pCopyDest, pCopySrc, rowsize, 
						FldHeight, dst_pitch*2, src_pitch*2);
		}
	}
	if (Use_Vertical_Filter)
	{
		if (SSEMMXenabled)
		{
			AvisynthTomsMoComp_SSE_VF_YV12(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
		else if (_3DNOWenabled)
		{
			AvisynthTomsMoComp_3DNOW_VF_YV12(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
		else
		{
			AvisynthTomsMoComp_MMX_VF_YV12(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
	}
	else
	{
		if (SSEMMXenabled)
		{
			AvisynthTomsMoComp_SSE_YV12(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
		else if (_3DNOWenabled)
		{
			AvisynthTomsMoComp_3DNOW_YV12(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
		else
		{
			AvisynthTomsMoComp_MMX_YV12(SearchEffortW, src_pitch, dst_pitch,
				rowsize, pWeaveSrc, pWeaveSrcP,
				pWeaveDest, TopFirst, pCopySrc,
				pCopySrcP, FldHeight);
		}
	}
	return 0;
}





int TomsMoComp::Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch)
{
BYTE* pDest = (BYTE*) dest;
BYTE* pSrc = (BYTE*) src;
	
	for (int i=0; i < rows; i++)
	{
		memcpy(pDest, pSrc, count);
		pSrc += src_pitch;
		pDest += dst_pitch;
	}
	return 0;
}


AVSValue __cdecl Create_TomsMoComp(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
    return new TomsMoComp(args[0].AsClip(), args[1].AsInt(), args[2].AsInt(),
		args[3].AsInt(), env);
}


extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("TomsMoComp", "ciii", Create_TomsMoComp, 0);
    return "`TomsMoComp' DeInterlace";
}

